use nom::combinator::not;

use crate::parser::error::CustomError;
use crate::parser::error::MyError;

use super::error::Res;
use super::object::PlainText;
use super::Context;

#[tracing::instrument(ret, level = "debug")]
pub fn plain_text<'r, 's>(context: Context<'r, 's>, input: &'s str) -> Res<&'s str, PlainText<'s>> {
    if input.len() == 0 {
        return Err(nom::Err::Error(CustomError::MyError(MyError(
            "Zero input length to plain_text.",
        ))));
    }
    let mut current_input = input.char_indices();
    loop {
        match current_input.next() {
            Some((offset, _char)) => {
                let remaining = &input[offset..];
                let exit_matcher_status = not(|i| context.check_exit_matcher(i))(remaining);
                if exit_matcher_status.is_err() {
                    if offset == 0 {
                        // If we're at the start of the input, then nothing is plain text, so fire an error for zero-length match.
                        exit_matcher_status?;
                    } else {
                        return Ok((
                            &input[offset..],
                            PlainText {
                                source: &input[..offset],
                            },
                        ));
                    }
                }
            }
            None => {
                // We hit the end of the file, so all input must be plain text
                return Ok((&input[input.len()..], PlainText { source: input }));
            }
        };
    }
}

#[cfg(test)]
mod tests {
    use nom::combinator::map;

    use crate::parser::object::Object;
    use crate::parser::parser_context::ContextElement;
    use crate::parser::parser_context::ContextTree;
    use crate::parser::parser_with_context::parser_with_context;
    use crate::parser::source::Source;

    use super::*;

    #[test]
    fn plain_text_simple() {
        let input = "foobarbaz";
        let initial_context: ContextTree<'_, '_> = ContextTree::new();
        let document_context =
            initial_context.with_additional_node(ContextElement::DocumentRoot(input));
        let plain_text_matcher = parser_with_context!(plain_text)(&document_context);
        let (remaining, result) = map(plain_text_matcher, Object::PlainText)(input).unwrap();
        assert_eq!(remaining, "");
        assert_eq!(result.get_source(), input);
    }
}
